/* cpu3 assembler A3  version 1.0

Prabhas Chongstitvatana  7th September 1999

format  (case insensitive)

.s                        define symbol
symbol value
 ...
.a address                set address
.b sym sym ...            define bytes
.c                        code segment
:label opcode operand
 ...
.e                        end

operand = #sym sym +sym
sym = numeric symbol label
numeric begins with 0..9
symbol  found in the symbol table

.s .a .b .c  can occur in any sequence
.e  is the last line of program

fix bug : undef op, add listing 2 Oct 1999

*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define	MXBUF	1024		/* max input line length */
#define	MXOP	33		/* number of opcode */
#define MXMEM	1000            /* size of memory */
#define INA	12
#define OUT	13
#define LDX	21

char	lbuf[MXBUF];		   /* for listing */
char 	ibuf[MXBUF], *cp = NULL;   /* input buffer, ch ptr */
static 	char sep[] = " \t\n";      /* separator char */
char    *w;
int 	lineno,loc;		   /* line num, current ads */
FILE 	*fi;
char 	mem[MXMEM];

typedef unsigned char uchar;

			/* opn 0 no operand, 1 has operand */
typedef struct { char name[5]; char opn; } optype;
optype	optable[] =
	{"ADD",1, "ADC",1, "COM",0, "AND",1, "ORA",1, "XOR",1,
	 "CLA",0, "LDA",1, "STA",1, "PSH",0, "POP",0, "NOP",0,
	 "INA",1, "OUT",1, "JZ",1,  "JC",1,  "JN",1,  "JMP",1,
	 "JSR",1, "RTS",0, "RTI",0, "LDX",1, "STX",1, "INX",0,
	 "DEX",0, "TSX",0, "ROL",0, "CLC",0, "STC",0, "CLI",0,
	 "STI",0, "ROR",0, "ID",0   };


error(char *s)
{
  printf("line %d error : %s symbol %s\n%s",lineno,s,w,lbuf);
  exit(0);
}

char *tok()	/* get one token from input */
{
  if( cp != NULL ) cp = strtok(NULL,sep);
  while ( cp == NULL ) {
    if( fgets(ibuf,MXBUF,fi) == NULL ) return NULL;
    lineno++;
    strcpy(lbuf,ibuf);
    cp = strtok(ibuf, sep);
  }
  cp = strupr(cp);   /* everything convert to uppercase */
  return cp;
}

int iscmd( char *w )	/* Is token a command ? */
{
  return ( w[0] == '.' );
}

twopass(int pass)   /* two-pass assembler; pass = 1,2 */
{
  int v, op, s, m;
  int mark;
			/* only getvalue() on pass2 */
  w = tok();
  while( strcmp(w,".E") != 0 ) {
    if( ! iscmd(w)) error("expect .");
    switch( w[1] ) {
    case 'S':		/* pass1  scan symbol into table */
      while(1) {        /* pass2  skip */
	w = tok();
	if( iscmd(w) ) break;
	if( pass == 1 ) putsym(w, atoi(tok()) );
      }
      break;
    case 'A' :          /* set address */
      w = tok();
      loc = atoi(w);
      w = tok();
      break;
    case 'B' :          /* pass1 skip */
      while(1) {        /* pass2 gen value */
	w = tok();
	if( iscmd(w) ) break;
	if( pass == 2) put1(getvalue(w));
	else loc++;
      }
      break;
    case 'C' :         /* pass1 collect label */
      while(1) {       /* pass2 gen actual code */
	w = tok();
	if( iscmd(w) ) break;
	if( w[0] == ':' )  {
	  if( pass == 1 ) {
	    putsym(w+1, loc);
	  }
	}else {
	  mark = loc;
	  op = getsym(w);
	  if( op < 0 ) error("undefined op");
	  if( isop0(op) )
	    put1(op << 2);
	  else {
	    w = tok();
	    s = sizeopr(w,op);
	    if(pass == 2) v = getvalue(w);
	    else v = 0;
	    m = getmode(w);
	    put1((op << 2) | m);
	    if( s == 1 ) put1(v);
	    else put2(v);
	  }
	  if( pass == 2 ) dolisting(mark);
	}
      }
    }
  }
}

initoptable()	/* put reserved words into op table */
{
  int i;
  for(i=0; i<MXOP; i++)
    putsym(optable[i].name,i);
}

int sizeopr( char *w, int op )  /* size of operand 1, 2 bytes */
{
  switch( w[0] ) {
  case '#' : return ( op == LDX ) ? 2 : 1;
  case '+' : return 1;
  default  : return ( op == INA || op == OUT ) ? 1 : 2;
  }
}

int getvalue( char *w )
{
  char *s;
  int v;

  if( w[0] == '#' || w[0] == '+' ) s = w+1;
  else s = w;
  if( isdigit(s[0]) ) return atoi(s);
  v = getsym(s);
  if( v < 0 ) error("symbol not define");
  return v;
}

int getmode( char *w )
{
  switch( w[0] ) {
  case '#' : return 1;
  case '+' : return 3;
  default  : return 2;
  }
}

put1( int v )
{
  mem[loc++] = v & 255;
}

put2( int v )
{
  mem[loc++] = (v >> 8) & 255;
  mem[loc++] = v & 255;
}

int isop0( int c )
{
  return(  optable[c].opn == 0 );
}

dolisting(int pc)	/* generate one line listing at pc */
{
  printf("%4d : %3d", pc, (uchar)mem[pc]);   	/* pc :  opcode */
  switch ( loc - pc ) {                         /* 2 opr8, 3 opr16 */
  case 1 : printf("         "); break;
  case 2 : printf(" %-3d      ", (uchar)mem[pc+1]); break;
  case 3 : printf(" $%-5d  ",256*(uchar)mem[pc+1]+(uchar)mem[pc+2]);
  }
  printf("%s", lbuf);
}

/* make two new file names from source, name.lst and name.s */
makename( char *source, char *lst, char *hex )
{
  int n;
  n = strcspn(source,".");
  strncpy(lst,source,n);
  strcpy(lst+n,".lst");
  strncpy(hex,source,n);
  strcpy(hex+n,".s");
}

main(int argc, char *argv[])
{
  char flst[80], fs[80];

  if( argc < 2 ) {
    printf("usage : a3 inputfile\n");
    exit(0);
  }

  initoptable();
  fi = fopen(argv[1],"r");
  lineno = 0;
  loc = 0;
  twopass(1);
/*  dumpsymbol();  */
  rewind(fi);

  makename(argv[1], flst, fs);
  freopen(flst,"w",stdout);
  lineno = 0;
  loc = 0;
  twopass(2);
  fclose(fi);

  freopen(fs,"w",stdout);
  Swriter(mem,0,loc);
  Swriter(mem,0,0);
}
